/**
 * @file   mfcnet.c
 * 
 * <EN>
 * @brief  feature input plugin
 * </EN>
 * 
 * <JA>
 * @brief  ħϥץ饰
 * </JA>
 * 
 * @author Nakamura Hitoshi
 * @date   2009.7.24
 * 
 * $Revision: 1.0 $
 * 
 */

/**
 * Required for a file
 *   - get_plugin_info()
 *
 * Optional for a file
 *   - initialize()
 * 
 */

/**
 * feature input plugin functions
 * 
 * Required:
 *   - fvin_get_optname()
 *   - fvin_get_configuration()
 *   - fvin_standby()
 *   - fvin_open()
 *   - fvin_read()
 *   - fvin_close()
 *
 * Optional:
 *   - fvin_terminate()
 *   - fvin_pause()
 *   - fvin_resume()
 * 
 */

#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include <julius/juliuslib.h>
#include <time.h>	// SPEED_TEST
#include <sys/time.h>	// SPEED_TEST
#include "plugin_defs.h"
#include "plugin_module.h"

// For tcpip of mfcnet
static Recog *recog;
static int mfcnet_sd = -1;	///< Listen socket for adinserv
#ifdef __TH_TEST1
// static int mfcnet_asd = -1;	///< Accept socket for adinserv => recog->mfcclist->mfcnet_asd
// static boolean input_is_end = FALSE; // => fvin_read()Υѿ
// static pid_t child;		/* child process ID (0 if myself is child) */
// static boolean SourceInfoFlg = TRUE;	// SourceInfo Read Flag =>  recog->mfcclist->SourceInfoFlg
#else
static int mfcnet_asd = -1;	///< Accept socket for adinserv
static boolean input_is_end = FALSE;
static pid_t child;		/* child process ID (0 if myself is child) */
static boolean SourceInfoFlg = TRUE;	// SourceInfo Read Flag
#endif
static short vec_size = 0;
static int adinnet_port = 0;

//SourceInfo source_info;	// igarashi 2013.09

extern int module_sd;

#define PLUGIN_TITLE "Feature vector input plugin for Julius.(MFCNET)"
#define INPUT_OPT "mfcnet"

#ifdef SPEED_TEST
static double start_time = 0.0;
#endif

#ifdef __TH_TEST1
static int mfc_read(void* buf, int size, int mfcnet_asd, int *input_is_end);
#else
static int mfc_read(void* buf, int size);
#endif
static void source_info_param(Recog *recog, void *dummy);
static boolean opt_outcode(Jconf *jconf, char *arg[], int argnum);

extern boolean is_module_mode();

#ifdef SPEED_TEST
// ߻֤
static double gettimeofday_nowtime()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec * 1e-6;
}

// ߻򳫻ϻѿ˥åȤ
static double gettimeofday_reset()
{
  start_time = gettimeofday_nowtime();
}

// ߻ȳϻκʬϤ߻򳫻ϻѿ˥åȤ
static double gettimeofday_diffout_and_reset()
{
  printf("diff time = %10.30f\n", gettimeofday_nowtime() - start_time);
  gettimeofday_reset();
}

// ǧ̤Ф륿ߥ󥰤ǴؿCall
void result_best_str(char *result_str)
{
  gettimeofday_diffout_and_reset();
}
#endif


#ifdef SEMAPHOR_ON
/*
void on_exit_post_semaphor(int result, void* arg)
{
  sem_post(sem_sync_result);

#ifdef SEMAPHOR_DEBUG_PRINT
  printf("***[SEMAPHOR_DEBUG_PRINT]*** on_exit_post_semaphor()\n");
#endif
}
*/

void sigint_handler_unlink_semaphor(int s)
{
  sem_unlink(sem_name);

#ifdef SEMAPHOR_DEBUG_PRINT
  printf("***[SEMAPHOR_DEBUG_PRINT]*** sigint_handler_unlink_semaphor()\n");
#endif

  exit(0);
}

void seg_begin_wait_semaphor()
{
  sem_wait(sem_sync_result);

#ifdef SEMAPHOR_DEBUG_PRINT
  int sval;
  sem_getvalue(sem_sync_result, &sval);
  printf("***[SEMAPHOR_DEBUG_PRINT]*** seg_begin_wait_semaphor() value is %d.\n", sval);
#endif
}

void recog_end_post_semaphor()
{
  sem_post(sem_sync_result);

#ifdef SEMAPHOR_DEBUG_PRINT
  int sval;
  sem_getvalue(sem_sync_result, &sval);
  printf("***[SEMAPHOR_DEBUG_PRINT]*** recog_end_post_semaphor() value is %d.\n", sval);
#endif
}
#endif //SEMAPHOR_ON

/************************************************************************/

/** 
 * <EN>
 * @brief  Initialization at loading time (optional)
 * 
 * If defined, this will be called just before this plugin is loaded to Julius.
 * if this returns -1, the whole functions in this file will not be loaded.
 *
 * This function is OPTIONAL.
 * </EN>
 * <JA>
 * @brief  ɤ߹߻νǤա
 *
 * ưJulius Υץ饰ɤ߹ݤ˺ǽ˸ƤФ롥
 * -1 ֤ȡΥץ饰Τɤ߹ޤʤʤ롥
 * ¹ԲǽΥå˻Ȥ롥
 *
 * </JA>
 * 
 * 
 * @return 0 on success, -1 on failure.
 * 
 */
int
initialize()
{
  j_add_option("-outcode", 1, 1, "select info to output to the module: WLPSCwlps", opt_outcode);

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  int ret = 0;
  printf("***[MFCNET_DEBUG_PRINT]*** initialize() : [out]return=%d\n", ret);
  return ret;
#endif
#else
  return 0;
#endif
}

int
startup(void *data)
{
  recog = data;
#ifdef __TH_TEST1
  recog->mfcclist->mfcnet_asd = -1;
  recog->mfcclist->SourceInfoFlg = TRUE;
#endif

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** startup() : [in]*data=0x%x(address)\n", data);
#endif
#endif

  adinnet_port = recog->jconf->input.adinnet_port;
#ifdef NO_MASK_DATA
  vec_size = recog->amlist->hmminfo->opt.vec_size;
#else
  vec_size = recog->amlist->hmminfo->opt.vec_size * 2;  // ǥͤƥޥʬݤΰ٤*2
#endif

#ifdef SEMAPHOR_ON
#ifdef SEMAPHOR_DEBUG_PRINT
  printf("***[SEMAPHOR_DEBUG_PRINT]*** startup() : semaphor init\n");
#endif

  sprintf(sem_name, "sync_result_%d", getpid());
  sem_unlink(sem_name);
  sem_sync_result = sem_open(sem_name, O_CREAT, S_IRWXU, 1);
  if (sem_sync_result == SEM_FAILED) {
    printf("semaphor open error (%d)\n", errno);
  }
  else {
    int sval;
    sem_getvalue(sem_sync_result, &sval);
    printf("semaphor open succeded. value is %d.\n", sval);
  }
  //on_exit(on_exit_post_semaphor, 0);
  signal(SIGINT, sigint_handler_unlink_semaphor);
  signal(SIGABRT, sigint_handler_unlink_semaphor);
  signal(SIGALRM, sigint_handler_unlink_semaphor);
  signal(SIGBUS, sigint_handler_unlink_semaphor);
  signal(SIGFPE, sigint_handler_unlink_semaphor);
  signal(SIGHUP, sigint_handler_unlink_semaphor);
  signal(SIGILL, sigint_handler_unlink_semaphor);
  signal(SIGKILL, sigint_handler_unlink_semaphor);
  signal(SIGPIPE, sigint_handler_unlink_semaphor);
  signal(SIGQUIT, sigint_handler_unlink_semaphor);
  signal(SIGSEGV, sigint_handler_unlink_semaphor);
  signal(SIGTERM, sigint_handler_unlink_semaphor);

#endif //SEMAPHOR_ON

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** startup() : [out]adinnet_port=%d\n", adinnet_port);
  printf("***[MFCNET_DEBUG_PRINT]*** startup() : [out]vec_size=%d\n", vec_size);
#endif
#endif

  return 0;
}

/** 
 * <EN>
 * @brief  Get information of this plugin (required)
 *
 * This function should return informations of this plugin file.
 * The required info will be specified by opcode:
 *  - 0: return description string of this file into buf
 *
 * This will be called just after Julius find this file and after
 * initialize().
 *
 * @param opcode [in] requested operation code
 * @param buf [out] buffer to store the return string
 * @param buflen [in] maximum length of buf
 *
 * @return 0 on success, -1 on failure.  On failure, Julius will ignore this
 * plugin.
 * 
 * </EN>
 * <JA>
 * @brief  ץ饰ɬܡ
 *
 * Υץ饰˴ؤ֤Ϳ줿 opcode ˤäư롥
 *  - 0 ξ硤Υץ饰ե̾ΤͿ줿Хåե˳Ǽ
 *
 * δؿϡJulius Υץ饰ɤ߹ľ˸ƤФ롥
 * 
 * @param opcode [in] ׵ư ( 0 Τ߼)
 * @param buf [out] ͤǼХåե
 * @param buflen [in] buf κĹ
 * 
 * @return 顼 -1,  0 ֤顼Ȥ -1 ֤硤
 * Υץ饰Τɤ߹ޤʤ
 * </JA>
 * 
 */
int
get_plugin_info(int opcode, char *buf, int buflen)
{
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** get_plugin_info() : [in]opcode=%d\n", opcode);
  printf("***[MFCNET_DEBUG_PRINT]*** get_plugin_info() : [in]buflen=%d\n", buflen);
#endif
#endif
  switch(opcode) {
  case 0:
    /* plugin description string */
    strncpy(buf, PLUGIN_TITLE, buflen);
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** get_plugin_info() : [out]buf=%s\n", buf);
#endif
#endif
    break;
  }

  return 0;
}

/************************************************************************/
/************************************************************************/
/* Feature-vector input plugin functions */

/** 
 * <EN>
 * @brief  Return option string to select this input at option. (required)
 *
 * This function should return option string which should be specified
 * as an argument "-input" option, to be used on Julius.  The returning
 * string should not be the same with any existing value.
 *
 * This function will be called several times at option parsing at startup.
 *
 * @param buf [out] buffer to store the return string
 * @param buflen [in] maximum length of buf
 * </EN>
 * <JA>
 * @brief  ѤΥץʸ֤ɬܡ
 *
 * Υץ饰ϤȤ򤹤ݤˡ"-input" ץǻꤹ
 * ٤ʸǼ֤֤ʸϡƥˤǤˤΤ䡤
 * ¾Υץ饰󤬻ѤƤΤƱǤʤȡ
 * ʤ⤷Ʊä祷ƥ¦ͥ褵
 *
 * δؿϡưΥץϻ˲٤ƤФ롥
 *
 * @param buf [out] ͤǼ֤Хåե
 * @param buflen [in] buf κĹ
 * </JA>
 * 
 */
void
fvin_get_optname(char *buf, int buflen)
{
  strncpy(buf, INPUT_OPT, buflen);

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_optname() : [in]buflen=%d\n", buflen);
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_optname() : [out]*buf=%s\n", buf);
#endif
#endif
}

/** 
 * <EN>
 * @brief  Return configuration parameters for this input (required)
 * 
 * This function should return configuration parameters about the input.
 *
 * When opcode = 0, return the dimension (length) of input vector.
 * 
 * When opcode = 1, return the frame interval (time between frames) in
 * milliseconds.
 * 
 * When opcode = 2, parameter type code can be returned.  The code should
 * the same format as used in HTK parameter file header.  This is for
 * checking the input parameter type against acousitc model, and
 * you can disable the checking by returning "0xffff" to this opcode.
 * 
 * @param opcode [in] requested operation code
 * 
 * @return values required for the opcode as described.
 * </EN>
 * 
 * <JA>
 * @brief  ħ̤Υѥ᡼֤ɬܡ
 *
 * ϥץ饰JuliusϤħ̤˴ؤѥ᡼֤
 * Ϳ줿ʲ opcode Ȥˡ֤ͤ
 *
 * opcode = 0: ٥ȥμ
 * opcode = 1: ե졼ढλñ̡ߥá
 * opcode = 2: ѥ᡼η
 *
 * opcode = 2 Υѥ᡼ηϡǥħ̤Ȥη
 * å˻Ȥ롥ͤϡHTK ħ̥եΥإå
 * 󥳡ɤ줿֤ͤåԤʤϡ
 * 0xffff ֤ȡ
 *
 * @param opcode [in] ׵ư ( 0 Τ߼)
 * 
 * @return opcode Ȥ׵ᤵ줿֤ͤ
 * </JA>
 */
int
fvin_get_configuration(int opcode)
{
int ret = 0;

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_configuration() : [in]opcode=%d\n", opcode);
#endif
#endif
  switch(opcode) {
  case 0:		/* return number of elements in a vector */
    ret = vec_size;		// ǡΥ֤
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_configuration() : [out]VectSize=%d\n", ret);
#endif
#endif
    return ret;
  case 1:/* return msec per frame */
    ret = 10;		// ȤꤢdefaultΤޤ
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_configuration() : [out]OneFrameTime=%d\n", ret);
#endif
#endif
    return ret;
  case 2:/* return parameter type specification in HTK format */

    /* return 0xffff to disable checking */
    ret = 0x0009;
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_get_configuration() : [out]ParamType=0x%x\n", ret);
#endif
#endif
    return ret;
  }
}

/************************************************************************/

/**
 * <EN>
 * @brief  Initialize input device (required)
 *
 * This will be called only once at start up of Julius.  You can
 * check if the input file exists or prepare a socket for connection.
 *
 * If this function returns FALSE, Julius will exit.
 * 
 * JuliusLib: this function will be called at j_adin_init().
 *
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ǥХɬܡ
 *
 * δؿϵư˰ƤФ롥Ǥϥեν
 * åȤѰդȤäϤΤνԤΤ˻Ȥ
 *
 * FALSE ֤硤Julius Ͻλ롥
 * 
 * JuliusLib: δؿ j_adin_init() ǸƤФ롥
 *
 * @return  TRUEԻ FALSE ֤
 * </JA>
 */
boolean
fvin_standby()
{
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_standby() : [in]adinnet_port=%d\n", adinnet_port);
#endif
#endif

  if ((mfcnet_sd = ready_as_server(adinnet_port)) < 0) {	// adinnet_port:ݡֹ(-adport)
    jlog("mfcnet_standby: cannot ready for server\n");
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_standby() : [outn]return=FALSE\n");
#endif
#endif
    return FALSE;
  }

  if (is_module_mode() == TRUE){
    plugin_module_setup(recog, NULL);
    callback_add(recog, CALLBACK_STATUS_PARAM, source_info_param, NULL);
#ifdef PLUGIN_DATA_DEBUG_PRINT
    printf("***[PLUGIN_DATA_DEBUG_PRINT]*** fvin_standby() : callback_add(CALLBACK_STATUS_PARAM)\n");
#endif
  }
#ifdef SEMAPHOR_ON
  else {
    callback_add(recog, CALLBACK_EVENT_RECOGNITION_END, recog_end_post_semaphor, NULL);
  }
#endif //SEMAPHOR_ON

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_standby() : [outn]return=TRUE\n");
#endif
#endif
  return TRUE;

}

/**
 * <EN>
 * @brief  Open an input (required)
 *
 * This function should open a new input.  You may open a feature
 * vector file, or wait for connection at this function.
 *
 * If this function returns FALSE, Julius will exit recognition loop.
 * 
 * JuliusLib: this will be called at j_open_stream().
 * 
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  Ϥ򳫤ɬܡ
 *
 * Ϥ򿷵˳եΥץ󡤥ͥåȥ饤Ȥ
 * ³ʤɤϤǹԤ
 *
 * FALSE ֤ȤJulius ǧ롼פȴ롥
 * 
 * JuliusLib: δؿ j_open_stream() ǸƤФ롥
 * 
 * @return  TRUEԻ FALSE ֤
 * </JA>
 */
boolean
#ifdef __TH_TEST1
fvin_open(MFCCCalc *mfcc)
#else
fvin_open()
#endif
{
  int status;

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [in]mfcnet_sd=%d\n", mfcnet_sd);
#endif
#endif

#ifndef __TH_TEST1    /* #ifdef FORK_MFCNET */
    /***********************************/
    /*** server infinite loop here!! ***/
    /***********************************/
    for (;;) {
      /* wait connection */
      jlog("waiting connection...\n");
      mfcnet_asd = accept_from(mfcnet_sd);	// 饤׵Ԥ
      //jlog("###### fvin_open mfcnet_asd(%#x)\n", mfcnet_asd);
      jlog_flush();		// ҥץʬΥ˿ƥץΥϽϤƤ
      /* fork self */
      child = fork();
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
      printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]mfcnet_asd=%d\n", mfcnet_asd);
      printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]child=%d\n", child);
#endif
#endif
      if (child < 0) {		/* error */
        jlog("mfcnet_standby: fork failed\n");
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]return=FALSE\n");
#endif
#endif
        return FALSE;
      }
      /* child thread should handle this request */
      if (child == 0) {		/* child thread */
        break;			/* proceed */		// ҥץϽȴ
      } else {			/* parent thread */
        jlog("forked process [%d] handles this request\n", child);
        jlog_flush();
        close_socket(mfcnet_asd);		// ƥץϤsocketǤ
      }
      while (waitpid(-1, &status, WNOHANG) > 0);	// ҥץβ
    }
#else  /* ~FORK_MFCNET */
    jlog("waiting connection... (NO FORK)\n");
#ifdef __TH_TEST1
    mfcc->mfcnet_asd = accept_from(mfcnet_sd);
#else
    mfcnet_asd = accept_from(mfcnet_sd);
#endif
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
#ifdef __TH_TEST1
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]mfcnet_asd=%d\n", mfcc->mfcnet_asd);
#else
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]mfcnet_asd=%d\n", mfcnet_asd);
#endif
#endif
#endif
#endif /* FORK_MFCNET */

#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_open() : [out]return=TRUE\n");
#endif
#endif

#ifdef SEMAPHOR_ON
  if (is_module_mode == FALSE){
    seg_begin_wait_semaphor();
  }
#endif //SEMAPHOR_ON

#ifdef SPEED_TEST
  gettimeofday_reset();
#endif

  return TRUE;
}

pid_t fvin_read_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/**
 * <EN>
 * @brief  Read a vector from input (required)
 *
 * This will be called repeatedly at each frame, and the read vector
 * will be processed immediately, and then this function is called again.
 *
 * Return value of ADIN_EOF tells end of stream to Julius, which
 * causes Julius to finish current recognition and close stream.
 * ADIN_SEGMENT requests Julius to segment the current input.  The
 * current recognition will be stopped at this point, recognition
 * result will be output, and then Julius continues to the next input.
 * The behavior of ADIN_SEGMENT is similar to ADIN_EOF except that
 * ADIN_SEGMENT does not close/open input, but just stop and restart
 * the recognition.  At last, return value should be ADIN_ERROR on
 * error, in which Julius exits itself immediately.
 * 
 * @param vecbuf [out] store a vector obtained in this function
 * @param veclen [in] vector length
 * 
 * @return 0 on success, ADIN_EOF on end of stream, ADIN_SEGMENT to
 * request segmentation to Julius, or ADIN_ERROR on error.
 * </EN>
 * <JA>
 * @brief Ϥ٥ȥɤ߹ɬܡ
 *
 * δؿϤ٥ȥ򣱤Ĥɤ߹ࡥδؿ
 * ե졼ऴȤ˸ƤФ졤ɤ߹ޤ줿٥ȥϤΤȤǧ졤
 * ޤΥե졼ΥǡɤिˤδؿƤФ롥
 *
 * ϤüޤãȤADIN_EOF ֤ΤȤJulius ϸ
 * ǧλϤĤ롥
 *
 * ADIN_ERROR Ϥδؿǿʥ顼֤줬֤
 * 줿硤Julius Ϥξǰ۾ｪλ롥
 *
 * ADIN_SEGMENT ֤ȤǡJulius ˸ߤǧ򸽻Ƕڤ뤳
 * ׵᤹뤳ȤǤ롥ߤǧϤλǤäڤ졤
 * ޤǤǧ̤ꡦϤ줿ȡǧϤޤꤳ
 * ؿƤФ롥ADIN_SEGMENT  ADIN_EOF ưƤ뤬
 * ADIN_EOF  adin_close(), adin_open() ƤϤλ
 * ΤФơADIN_SEGMENT ϤƤФϤ³Ԥ롥
 * 
 * @param vecbuf [out] 줿٥ȥǼХåե
 * @param veclen [in] ٥ȥĹ
 * 
 * @return
 *  0                : 
 * -1(ADIN_EOF)      : Ϥüޤã
 * -2(ADIN_ERROR)    : 顼
 * -3(ADIN_SEGMENT)  : ߤǧ򸽻Ƕڤ׵Ф(ڤȤߤʤΤǤΥѥ᡼֤ʤ)
 * </JA>
 */
int
#ifdef __TH_TEST1
fvin_read(float *vecbuf, int veclen, MFCCCalc *mfcc)
#else
fvin_read(float *vecbuf, int veclen)
#endif
{
#ifdef __TH_TEST1
  int input_is_end = FALSE;
#endif
  int read_count = 0;
  float *mask;
#ifdef NO_MASK_DATA
  float tmp_buf[27];	// ޥǡбޤǤ˥ǡ
#endif
  int ret = 0;

#ifdef MFCNET_DEBUG_PRINT
  int i = 0;
  int i_tmp_buf = 0;
  pid_t pid = getpid();
#ifdef MFCNET_DEBUG_PARAM_PRINT
  if (fvin_read_dbgprint_pass != pid){
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [in]vecbuf=0x%x\n", vecbuf);
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [in]veclen=%d\n", veclen);
#ifdef __TH_TEST1
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [in]SourceInfoFlg=%d\n", mfcc->SourceInfoFlg);
#else
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [in]SourceInfoFlg=%d\n", SourceInfoFlg);
#endif
  }
#endif
#endif

#ifdef __TH_TEST1
  if (mfcc->SourceInfoFlg == TRUE){	//1ȯƬΤ
#else
  if (SourceInfoFlg == TRUE){	//1ȯƬΤ
#endif
    do {
#ifdef __TH_TEST1
//++      read_count = mfc_read((void*)&source_info, sizeof(SourceInfo), mfcc->mfcnet_asd, &input_is_end);
      read_count = mfc_read((void*)&mfcc->source_info, sizeof(SourceInfo), mfcc->mfcnet_asd, &input_is_end);
#else
      read_count = mfc_read((void*)&source_info, sizeof(SourceInfo));
#endif
    } while(read_count == 0);
//++    printf("source_id = %d, azimuth = %f, elevation = %f, sec = %lld, usec = %lld\n",
//++           source_info.source_id, source_info.azimuth, source_info.elevation,
//++           source_info.time.tv_sec, source_info.time.tv_usec);
    printf("source_id = %d, azimuth = %f, elevation = %f, sec = %lld, usec = %lld\n",
           mfcc->source_info.source_id, mfcc->source_info.azimuth, mfcc->source_info.elevation,
           mfcc->source_info.time.tv_sec, mfcc->source_info.time.tv_usec);
#ifdef __TH_TEST1
    mfcc->SourceInfoFlg = FALSE;
#else
    SourceInfoFlg = FALSE;
#endif
  }

  do {
    // ħ̥ǡ
#ifdef __TH_TEST1
#ifdef NO_MASK_DATA
    read_count = mfc_read(vecbuf, (veclen * sizeof(float)), mfcc->mfcnet_asd, &input_is_end);
#else
    read_count = mfc_read(vecbuf, (veclen * sizeof(float) / 2), mfcc->mfcnet_asd, &input_is_end);
#endif
#else
#ifdef NO_MASK_DATA
    read_count = mfc_read(vecbuf, (veclen * sizeof(float)));
#else
    read_count = mfc_read(vecbuf, (veclen * sizeof(float) / 2));
#endif
#endif

#ifdef MFCNET_DEBUG_PRINT
    if (fvin_read_dbgprint_pass != pid){
#ifdef MFCNET_DEBUG_PARAM_PRINT
      printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [out]read_count=%d\n", read_count);
#endif
#ifdef MFCNET_DEBUG_DATA_PRINT
      for (i = 0; i < read_count; i++){
        memcpy(&i_tmp_buf, &vecbuf[i], sizeof(int));
        printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [out]vecbuf[%d]=%x\n", i, i_tmp_buf);
      }
#endif
    }
#endif
  } while (read_count == 0);	// Ǥޤ(ȥ饤)
  if (read_count > 0) {
#ifdef NO_MASK_DATA
    mask = tmp_buf;	// ޥǡбޤǤ˥ǡ
#else
    mask = vecbuf + read_count;	//ޥ񤭹ߥХåե֤򤺤餹
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  if (fvin_read_dbgprint_pass != pid){
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : mask=0x%x(mask-vecbuf=%d)\n", mask, (int)mask-(int)vecbuf);
  }
#endif
#endif
#endif
    do {
      // ޥǡ
#ifdef __TH_TEST1
#ifdef NO_MASK_DATA
      read_count = mfc_read(mask, (veclen * sizeof(float)), mfcc->mfcnet_asd, &input_is_end);
#else
      read_count = mfc_read(mask, (veclen * sizeof(float) / 2), mfcc->mfcnet_asd, &input_is_end);
#endif
#else
#ifdef NO_MASK_DATA
      read_count = mfc_read(mask, (veclen * sizeof(float)));
#else
      read_count = mfc_read(mask, (veclen * sizeof(float) / 2));
#endif
#endif
      
#ifdef MFCNET_DEBUG_PRINT
      if (fvin_read_dbgprint_pass != pid){
#ifdef MFCNET_DEBUG_PARAM_PRINT
        printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [out]read_count=%d\n", read_count);
#endif
#ifdef MFCNET_DEBUG_DATA_PRINT
        for (i = 0; i < read_count; i++){
          printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [out]mask[%d]=%f\n", i, mask[i]);
        }
#endif
      }
#endif
    } while (read_count == 0);	// Ǥޤ(ȥ饤)
  }

  if (read_count < 0) {
    printf("read_count < 0, read_count=%d, veclen=%d\n", read_count, veclen);
  }

  /* read one vector from the input */
  switch(read_count){
  case -1:
    if (input_is_end == TRUE){	// rdؿˤɤ߹Хȿ0ʤΤϽü
      ret = ADIN_EOF;
    } else {
      //ret = ADIN_ERROR;	// HARKκǽǡ(int)0б餳ˤ
      ret = ADIN_EOF;		// HARKǽǡ(int)0֤ƤʤΤǡinput_is_end == TRUEפȤϤʤʤΤǤADIN_EOFƤ
    }
    break;
  case -2:
  case -3:
    ret = ADIN_ERROR;
    break;
  default :
    ret = 0;
    break;
  }

#ifdef MFCNET_DEBUG_PRINT
  if (fvin_read_dbgprint_pass != pid){
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** fvin_read() : [out]return=%d (0:OK, %d:ADIN_EOF, %d:ADIN_ERROR, %d:ADIN_SEGMENT)\n"
           , ret, ADIN_EOF, ADIN_ERROR, ADIN_SEGMENT);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    fvin_read_dbgprint_pass = pid;
#endif
  }
#endif

  return ret;
}

/**
 * <EN>
 * @brief  Close the current input (required)
 *
 * This function will be called when the input has reached end of file
 * (i.e. the last call of fvin_read() returns ADIN_EOF)
 *       
 * You may close a file or disconnect network client here.
 *
 * If this function returns TRUE, Julius will go again to adin_open()
 * to open another stream.  If returns FALSE, Julius will exit
 * the recognition loop.
 * 
 * JuliusLib: This will be called at the end of j_recognize_stream().
 * 
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ϤĤɬܡ
 *
 * ߤϤĤ롥δؿϡϤüEOFˤãȤʤ
 *  fvin_read()  ADIN_EOF ֤Ȥˤ˸ƤФ롥̾Ǥ
 * եĤ롤ͥåȥ³ǤʤɤνԤ
 *
 * ｪλȤTRUE֤ȤJulius  adin_open() ä
 * ¾Υȥ꡼򳫤Ȥ롥 FALSE ֤ȤϡJulius 
 * ǧ롼פȴ롥
 * 
 * JuliusLib: δؿ j_recognize_stream() κǸǸƤФ롥
 * 
 * @return  TRUEԻ FALSE ֤
 * </JA>
 */
boolean
#ifdef __TH_TEST1
fvin_close(MFCCCalc *mfcc)
#else
fvin_close()
#endif
{
  /* end of connection */
#ifdef __TH_TEST1
  close_socket(mfcc->mfcnet_asd);
#else
  close_socket(mfcnet_asd);
#endif

  jlog("connection end\n");

#ifdef __TH_TEST1
  mfcc->SourceInfoFlg = TRUE;
#else
  SourceInfoFlg = TRUE;
#endif
#ifdef __TH_TEST1
  return TRUE; // NO FORK ǥƥ
#else
  return FALSE;	// ǧ롼פȴΤFALSE
#endif
}

/************************************************************************/

/**
 * <EN>
 * @brief  A hook for Termination request (optional)
 *
 * This function will be called when Julius receives a Termination
 * request to stop running.  This can be used to synchronize input
 * facility with Julius's running status.
 * 
 * Termination will occur when Julius is running on module mode and
 * received TERMINATE command from client, or j_request_terminate()
 * is called inside application.  On termination, Julius will stop
 * recognition immediately (discard current input if in process),
 * and wait until received RESUME command or call of j_request_resume().
 *
 * This hook function will be called just after a Termination request.
 * Please note that this will be called when Julius receives request,
 * not on actual termination.
 * 
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ׵ѥեåǤա
 *
 * δؿȡJulius ׵äݤˤδؿƤӽФ
 * ȤäơJulius ǡƳƱƱ뤳Ȥ
 * Ǥ롥㡧Ф׵Фʤɡ
 *
 * ׵ϡJulius ץꥱ䥯饤Ȥ
 * ǧǤ׵Ǥ롥ŪˤϡJulius ⥸塼⡼ɤư
 *  TERMINATE ޥɤ򥯥饤ȤäȤ䡤
 * JuliusLibȤ߹ץꥱ j_request_terminate() 
 * ƤȤȯ롥
 *
 * ׵ȡJulius ϸߤǧǤ롥
 * ǧǤä硤Ϥ˴¨Ǥ롥
 * κƳϡRESUME ޥɤ j_request_resume() θƤӽФǹԤ롥
 *
 * δؿ׵ Julius äǸƤФ롥
 * ºݤ˽ǤǸƤФΤǤϤʤȤդ줿
 * 
 * @return  TRUE, 顼 FALSE ֤
 * </JA>
 * 
 */
boolean
fvin_terminate()
{
  printf("terminate request\n");
  return TRUE;
}

/**
 * <EN>
 * @brief  A hook for Pause request (optional)
 *
 * This function will be called when Julius receives a Pause request
 * to stop running.  This can be used to synchronize input facility
 * with Julius's running status.
 * 
 * Pause will occur when Julius is running on module mode and
 * received PAUSE command from client, or j_request_pause()
 * is called inside application.  On pausing, Julius will 
 * stop recognition and then wait until it receives RESUME command
 * or j_request_resume() is called.  When pausing occurs while recognition is
 * running, Julius will process it to the end before stops.
 *
 * This hook function will be called just after a Pause request.
 * Please note that this will be called when Julius receives request,
 * not on actual pause.
 *
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ׵ѥեåǤա
 *
 * δؿȡJulius ׵äݤˤδؿƤӽФ
 * ȤäơJulius ǡƳƱƱ뤳Ȥ
 * Ǥ롥㡧Ф׵Фʤɡ
 *
 * ׵ϡJulius ץꥱ䥯饤Ȥ롤
 * ǧΰߤ׵Ǥ롥ŪˤϡJulius ⥸塼⡼ɤư
 *  PAUSE ޥɤ򥯥饤ȤäȤ䡤
 * JuliusLibȤ߹ץꥱ j_request_pause() 
 * ƤȤȯ롥
 *
 * ׵ȡJulius ϸߤǧǤ롥
 * ǧǤä硤ǧޤԤäƤǤ롥
 * κƳϡRESUME ޥɤ j_request_resume() θƤӽФǹԤ롥
 * 
 * ׵ (fvin_terminate) Ȥΰ㤤ϡǧ׵Ȥư
 * ۤʤ롥׵Ǥ϶Ǥ뤬׵ǤϤǧޤ
 * ԤäƤߤ롥
 *
 * δؿ׵ Julius äǸƤФ롥
 * ºݤ˽ߤǸƤФΤǤϤʤȤդ줿
 * 
 * @return  TRUE, 顼 FALSE ֤
 * </JA>
 * 
 */
boolean
fvin_pause()
{
  printf("pause request\n");
  return TRUE;
}

/**
 * <EN>
 * @brief  A hook for Resume request (optional)
 *
 * This function will be called when Julius received a resume request
 * to recover from pause/termination status.
 * 
 * Resume will occur when Julius has been stopped by receiving RESUME
 * command from client on module mode, or j_request_resume() is called
 * inside application.
 *
 * This hook function will be called just after a resume request.
 * This can be used to make this A/D-in plugin cooperate with the
 * pause/resume status, for example to tell audio client to restart
 * audio streaming.
 *
 * This function is totally optional.
 * 
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ǧƳ׵ѥեåǤա
 *
 * δؿȡJulius ߾֤ǧƳ׵κݤ
 * δؿƤӽФ
 *
 * ǧƳ׵ϡJulius ⥸塼⡼ɤư RESUME ޥɤ
 * 饤ȤäȤ䡤JuliusLibȤ߹ץꥱ
 *  j_request_resume() ƤȤȯ롥κƳ׵᤬ȯ
 * ȡJulius ߤƤǧƳ롥
 *
 * աδؿϡºݤߤȤ˸ƤФΤǤϤʤJulius 
 * ׵äǡΤӤ˸ƤФ롥ʣƤФ뤳Ȥ䡤
 * ǤưǤˤˤΥޥɤäȤˤƤ
 * 뤳Ȥ뤳Ȥդ줿
 * 
 * @return  TRUE, 顼 FALSE ֤
 * </JA>
 * 
 */
boolean
fvin_resume()
{
  printf("resume request\n");
  return TRUE;
}

pid_t mfc_read_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
static int
#ifdef __TH_TEST1
mfc_read(void* buf, int size, int mfcnet_asd, int *input_is_end)
#else
mfc_read(void* buf, int size)
#endif
{
  int cnt, ret, ref;
  fd_set rfds;
  struct timeval tv;
  int status;

#ifdef MFCNET_DEBUG_PRINT
  pid_t pid = getpid();
#ifdef MFCNET_DEBUG_PARAM_PRINT
  if (mfc_read_dbgprint_pass != pid){
    printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [in]size=%d\n", size);
  }
#endif
#endif

  /* check if some commands are waiting in queue */
  FD_ZERO(&rfds);		// fd_set(SOCKETط)
  FD_SET(mfcnet_asd, &rfds);	// mfcnet_asd(SOCKETط)
  tv.tv_sec = 0;
  tv.tv_usec = 1;
  status = select(mfcnet_asd+1, &rfds, NULL, NULL, &tv);	// ǡԤ(SOCKETط)
  if (status < 0) {		/* error */
    jlog("adin_tcpip_read: cannot poll\n");
    ref = -2;
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]return=%d\n", ref);
#endif
#endif
    return ref;			/* error return */
  }
  if (status > 0) {		/* there are some data */
    /* read one data segment, leave rest even if any for avoid blocking */
    ret = rd(mfcnet_asd, (char *)buf, &cnt, size);
    if (ret == 0) {
#ifdef __TH_TEST1
      *input_is_end = TRUE;
#else
      input_is_end = TRUE;
#endif
      ref = -1;
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]input_is_end=TRUE\n");
  printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]return=%d\n", ref);
#endif
#endif
      return ref;
    }
    if (ret < 0) {
      /* error */
#ifdef __TH_TEST1
      *input_is_end = FALSE;
#else
      input_is_end = FALSE;
#endif
      ref = -1;
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
  printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]input_is_end=FALSE\n");
  printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]return=%d\n", ref);
#endif
#endif
      return ref;
    }

    if (cnt != size){
      jlog("mfc_read() : cnt(%d) <> size(%d)\n", cnt, size);
      ref = -3;
#ifdef MFCNET_DEBUG_PRINT
#ifdef MFCNET_DEBUG_PARAM_PRINT
      printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]return=%d\n", ref);
#endif
#endif
      return ref;// ǡ԰
    }

  } else {	/* no data */
    cnt = 0;
  }
  cnt /= sizeof(float);

#ifdef MFCNET_DEBUG_PRINT
  if (mfc_read_dbgprint_pass != pid){
#ifdef MFCNET_DEBUG_PARAM_PRINT
    printf("***[MFCNET_DEBUG_PRINT]*** mfc_read() : [out]return=cnt=%d\n", cnt);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    mfc_read_dbgprint_pass = pid;
#endif
  }
#endif

  return cnt;
}

static void
source_info_param(Recog *recog, void *dummy)
{
  char buf[256];

#ifdef PLUGIN_DATA_DEBUG_PRINT
    printf("***[PLUGIN_DATA_DEBUG_PRINT]*** source_info_param() : [in]\n");
#endif
//++  sprintf(buf, "<SOURCEINFO SOURCEID=\"%d\" AZIMUTH=\"%f\" ELEVATION=\"%f\" SEC=\"%lld\" USEC=\"%lld\"/>\n.\n",
//++		  source_info.source_id, source_info.azimuth, source_info.elevation,			// igarashi 2013.09 Original
//++          source_info.time.tv_sec, source_info.time.tv_usec);
//  sprintf(buf, "-------- SOURCEID = %d --------\n", recog->mfcclist->source_info.source_id);
  sprintf(buf, "<SOURCEINFO SOURCEID=\"%d\" AZIMUTH=\"%f\" ELEVATION=\"%f\" SEC=\"%lld\" USEC=\"%lld\"/>\n.\n",
		  recog->mfcclist->source_info.source_id, recog->mfcclist->source_info.azimuth, recog->mfcclist->source_info.elevation,
          recog->mfcclist->source_info.time.tv_sec, recog->mfcclist->source_info.time.tv_usec);
#ifdef SEMAPHOR_ON
  seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.101[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  module_send(module_sd, buf);
#ifdef SEMAPHOR_ON
  recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.101[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
}

static boolean
opt_outcode(Jconf *jconf, char *arg[], int argnum)
{
  plugin_decode_output_selection(arg[0]);
  return TRUE;
}

/* end of file */
